@@ -52,7 +52,8 @@
 #define DEFAULT_SERVER_NAME "127.0.0.1"
 #define DEFAULT_SERVER_PORT 53
 #define DEFAULT_SERVER_DOT_PORT 853
-#define DEFAULT_SERVER_PORTS "udp/tcp 53 or DoT 853"
+#define DEFAULT_SERVER_DOH_PORT 443
+#define DEFAULT_SERVER_PORTS "udp/tcp 53, DoT 853 or DoH 443"
 #define DEFAULT_LOCAL_PORT 0
 #define DEFAULT_MAX_OUTSTANDING 100
 #define DEFAULT_TIMEOUT 5
@@ -87,6 +88,7 @@ typedef struct {
     bool               updates;
     bool               verbose;
     enum perf_net_mode mode;
+    perf_suppress_t    suppress;
 } config_t;
 
 typedef struct {
@@ -398,15 +400,18 @@ stringify(unsigned int value)
 static void
 setup(int argc, char** argv, config_t* config)
 {
-    const char* family      = NULL;
-    const char* server_name = DEFAULT_SERVER_NAME;
-    in_port_t   server_port = 0;
-    const char* local_name  = NULL;
-    in_port_t   local_port  = DEFAULT_LOCAL_PORT;
-    const char* filename    = NULL;
-    const char* edns_option = NULL;
-    const char* tsigkey     = NULL;
-    const char* mode        = 0;
+    const char* family         = NULL;
+    const char* server_name    = DEFAULT_SERVER_NAME;
+    in_port_t   server_port    = 0;
+    const char* local_name     = NULL;
+    in_port_t   local_port     = DEFAULT_LOCAL_PORT;
+    const char* filename       = NULL;
+    const char* edns_option    = NULL;
+    const char* tsigkey        = NULL;
+    const char* mode           = 0;
+    const char* doh_uri        = DEFAULT_DOH_URI;
+    const char* doh_method     = DEFAULT_DOH_METHOD;
+    const char* local_suppress = 0;
 
     memset(config, 0, sizeof(*config));
     config->argc = argc;
@@ -422,7 +427,7 @@ setup(int argc, char** argv, config_t* config)
     perf_opt_add('f', perf_opt_string, "family",
         "address family of DNS transport, inet or inet6", "any",
         &family);
-    perf_opt_add('m', perf_opt_string, "mode", "set transport mode: udp, tcp or dot", "udp", &mode);
+    perf_opt_add('m', perf_opt_string, "mode", "set transport mode: udp, tcp, dot or doh", "udp", &mode);
     perf_opt_add('s', perf_opt_string, "server_addr",
         "the server to query", DEFAULT_SERVER_NAME, &server_name);
     perf_opt_add('p', perf_opt_port, "port",
@@ -480,6 +485,13 @@ setup(int argc, char** argv, config_t* config)
     perf_opt_add('v', perf_opt_boolean, NULL,
         "verbose: report each query and additional information to stdout",
         NULL, &config->verbose);
+    perf_long_opt_add("doh-uri", perf_opt_string, "doh_uri",
+        "the URI to use for DNS-over-HTTPS", DEFAULT_DOH_URI, &doh_uri);
+    perf_long_opt_add("doh-method", perf_opt_string, "doh_method",
+        "the HTTP method to use for DNS-over-HTTPS: GET or POST", DEFAULT_DOH_METHOD, &doh_method);
+    perf_long_opt_add("suppress", perf_opt_string, "message[,message,...]",
+        "suppress messages/warnings, see man-page for list of message types", NULL, &local_suppress);
+
     bool log_stdout = false;
     perf_opt_add('W', perf_opt_boolean, NULL, "log warnings and errors to stdout instead of stderr", NULL, &log_stdout);
 
@@ -489,13 +501,33 @@ setup(int argc, char** argv, config_t* config)
         perf_log_tostdout();
     }
 
+    config->suppress = perf_opt_parse_suppress(local_suppress);
+
     if (mode != 0)
         config->mode = perf_net_parsemode(mode);
 
     if (!server_port) {
-        server_port = config->mode == sock_dot ? DEFAULT_SERVER_DOT_PORT : DEFAULT_SERVER_PORT;
+        switch (config->mode) {
+        case sock_doh:
+            server_port = DEFAULT_SERVER_DOH_PORT;
+            break;
+        case sock_dot:
+            server_port = DEFAULT_SERVER_DOT_PORT;
+            break;
+        default:
+            server_port = DEFAULT_SERVER_PORT;
+            break;
+        }
     }
 
+    if (doh_uri) {
+        perf_net_doh_parse_uri(doh_uri);
+    }
+    if (doh_method) {
+        perf_net_doh_parse_method(doh_method);
+    }
+    perf_net_doh_set_max_concurrent_streams(config->max_outstanding);
+
     if (family != NULL)
         config->family = perf_net_parsefamily(family);
     perf_net_parseserver(config->family, server_name, server_port,
@@ -669,19 +701,14 @@ do_send(void* arg)
             q->sock = tinfo->socks[tinfo->current_sock++ % tinfo->nsocks];
             switch (perf_net_sockready(q->sock, threadpipe[0], TIMEOUT_CHECK_TIME)) {
             case 0:
-                if (config->verbose) {
+                if (config->verbose && !config->suppress.sockready) {
                     perf_log_warning("socket %p not ready", q->sock);
                 }
                 q->sock  = 0;
                 all_fail = false;
                 continue;
             case -1:
-                if (errno == EINPROGRESS) {
-                    any_inprogress = 1;
-                    q->sock        = 0;
-                    continue;
-                }
-                if (config->verbose) {
+                if (config->verbose && !config->suppress.sockready) {
                     perf_log_warning("socket %p readiness check timed out", q->sock);
                 }
                 q->sock = 0;
@@ -744,12 +771,12 @@ do_send(void* arg)
             config->server_addr.length);
         if (n < 0) {
             if (errno == EINPROGRESS) {
-                if (config->verbose) {
+                if (config->verbose && !config->suppress.congestion) {
                     perf_log_warning("network congested, packet sending in progress");
                 }
                 any_inprogress = 1;
             } else {
-                if (config->verbose) {
+                if (config->verbose && !config->suppress.sendfailed) {
                     char __s[256];
                     perf_log_warning("failed to send packet: %s", perf_strerror_r(errno, __s, sizeof(__s)));
                 }
@@ -759,7 +786,9 @@ do_send(void* arg)
                 continue;
             }
         } else if ((unsigned int)n != length) {
-            perf_log_warning("failed to send full packet: only sent %d of %u", n, length);
+            if (!config->suppress.sendfailed) {
+                perf_log_warning("failed to send full packet: only sent %d of %u", n, length);
+            }
             PERF_LOCK(&tinfo->lock);
             query_move(tinfo, q, prepend_unused);
             PERF_UNLOCK(&tinfo->lock);
@@ -773,7 +802,7 @@ do_send(void* arg)
     while (any_inprogress) {
         any_inprogress = 0;
         for (i = 0; i < tinfo->nsocks; i++) {
-            if (perf_net_sockready(tinfo->socks[i], threadpipe[0], TIMEOUT_CHECK_TIME) == -1 && errno == EINPROGRESS) {
+            if (!perf_net_sockready(tinfo->socks[i], threadpipe[0], TIMEOUT_CHECK_TIME)) {
                 any_inprogress = 1;
             }
         }
@@ -806,12 +835,14 @@ process_timeouts(threadinfo_t* tinfo, uint64_t now)
 
         tinfo->stats.num_timedout++;
 
-        if (q->desc != NULL) {
-            perf_log_printf("> T %s", q->desc);
-        } else {
-            perf_log_printf("[Timeout] %s timed out: msg id %u",
-                config->updates ? "Update" : "Query",
-                (unsigned int)(q - tinfo->queries));
+        if (!config->suppress.timeouts) {
+            if (q->desc != NULL) {
+                perf_log_printf("> T %s", q->desc);
+            } else {
+                perf_log_printf("[Timeout] %s timed out: msg id %u",
+                    config->updates ? "Update" : "Query",
+                    (unsigned int)(q - tinfo->queries));
+            }
         }
         q = perf_list_tail(tinfo->outstanding_queries);
     } while (q != NULL && q->timestamp < now && now - q->timestamp >= config->timeout);
@@ -1115,9 +1146,8 @@ static void perf__net_event(struct perf_net_socket* sock, perf_socket_event_t ev
     stats_t* stats = &((threadinfo_t*)sock->data)->stats;
 
     switch (event) {
-    case perf_socket_event_reconnect:
-        stats->num_conn_reconnect++;
-    case perf_socket_event_connect:
+    case perf_socket_event_reconnected:
+    case perf_socket_event_connected:
         stats->num_conn_completed++;
 
         stats->conn_latency_sum += elapsed_time;
@@ -1126,6 +1156,14 @@ static void perf__net_event(struct perf_net_socket* sock, perf_socket_event_t ev
             stats->conn_latency_min = elapsed_time;
         if (elapsed_time > stats->conn_latency_max)
             stats->conn_latency_max = elapsed_time;
+        break;
+
+    case perf_socket_event_reconnecting:
+        stats->num_conn_reconnect++;
+        break;
+
+    default:
+        break;
     }
 }
 
@@ -1179,13 +1217,11 @@ threadinfo_init(threadinfo_t* tinfo, const config_t* config,
         tinfo->socks[i] = perf_net_opensocket(config->mode, &config->server_addr,
             &config->local_addr,
             socket_offset++,
-            config->bufsize);
+            config->bufsize,
+            tinfo, perf__net_sent, perf__net_event);
         if (!tinfo->socks[i]) {
             perf_log_fatal("perf_net_opensocket(): no socket returned, out of memory?");
         }
-        tinfo->socks[i]->data  = tinfo;
-        tinfo->socks[i]->sent  = perf__net_sent;
-        tinfo->socks[i]->event = perf__net_event;
     }
     tinfo->current_sock = 0;
 
@@ -1202,14 +1238,16 @@ threadinfo_stop(threadinfo_t* tinfo)
 }
 
 static void
-threadinfo_cleanup(threadinfo_t* tinfo, times_t* times)
+threadinfo_cleanup(config_t* config, threadinfo_t* tinfo, times_t* times)
 {
     unsigned int i;
 
     if (interrupted)
         cancel_queries(tinfo);
-    for (i = 0; i < tinfo->nsocks; i++)
+    for (i = 0; i < tinfo->nsocks; i++) {
+        perf_net_stats_compile(config->mode, tinfo->socks[i]);
         perf_net_close(tinfo->socks[i]);
+    }
     if (tinfo->last_recv > times->end_time)
         times->end_time = tinfo->last_recv;
 }
@@ -1244,6 +1282,7 @@ int main(int argc, char** argv)
     switch (config.mode) {
     case sock_tcp:
     case sock_dot:
+    case sock_doh:
         // block SIGPIPE for TCP/DOT mode, if connection is closed it will generate a signal
         perf_os_blocksignal(SIGPIPE, true);
         break;
@@ -1294,13 +1333,16 @@ int main(int argc, char** argv)
     if (config.stats_interval > 0)
         PERF_JOIN(stats_thread.sender, NULL);
 
+    perf_net_stats_init(config.mode);
+
     for (i = 0; i < config.threads; i++)
-        threadinfo_cleanup(&threads[i], &times);
+        threadinfo_cleanup(&config, &threads[i], &times);
 
     print_final_status(&config);
 
     sum_stats(&config, &total_stats);
     print_statistics(&config, &times, &total_stats);
+    perf_net_stats_print(config.mode);
 
     cleanup(&config);
 #if OPENSSL_VERSION_NUMBER < 0x10100000L
