@@ -53,7 +53,8 @@
 #define DEFAULT_SERVER_NAME "127.0.0.1"
 #define DEFAULT_SERVER_PORT 53
 #define DEFAULT_SERVER_DOT_PORT 853
-#define DEFAULT_SERVER_PORTS "udp/tcp 53 or DoT 853"
+#define DEFAULT_SERVER_DOH_PORT 443
+#define DEFAULT_SERVER_PORTS "udp/tcp 53, DoT 853 or DoH 443"
 #define DEFAULT_LOCAL_PORT 0
 #define DEFAULT_SOCKET_BUFFER 32
 #define DEFAULT_TIMEOUT 45
@@ -103,6 +104,8 @@ static uint64_t query_timeout;
 static bool     edns;
 static bool     dnssec;
 
+static perf_ednsoption_t* edns_option = 0;
+
 static perf_datafile_t* input;
 
 /* The target traffic level at the end of the ramp-up */
@@ -202,6 +205,8 @@ static perf_tsigkey_t* tsigkey;
 static bool         verbose;
 static unsigned int max_fall_behind;
 
+static perf_suppress_t suppress;
+
 const char* progname = "resperf";
 
 static char*
@@ -216,8 +221,23 @@ stringify(double value, int precision)
 static void perf__net_event(struct perf_net_socket* sock, perf_socket_event_t event, uint64_t elapsed_time);
 static void perf__net_sent(struct perf_net_socket* sock, uint16_t qid);
 
-static void
-setup(int argc, char** argv)
+static ramp_bucket* init_buckets(int n)
+{
+    ramp_bucket* p;
+    int          i;
+
+    if (!(p = calloc(n, sizeof(*p)))) {
+        perf_log_fatal("out of memory");
+        return 0; // fix clang scan-build
+    }
+    for (i = 0; i < n; i++) {
+        p[i].queries = p[i].responses = p[i].failures = 0;
+        p[i].latency_sum                              = 0.0;
+    }
+    return p;
+}
+
+static void setup(int argc, char** argv)
 {
     const char*  family      = NULL;
     const char*  server_name = DEFAULT_SERVER_NAME;
@@ -229,7 +249,11 @@ setup(int argc, char** argv)
     int          sock_family;
     unsigned int bufsize;
     unsigned int i;
-    const char*  _mode = 0;
+    const char*  _mode           = 0;
+    const char*  edns_option_str = NULL;
+    const char*  doh_uri         = DEFAULT_DOH_URI;
+    const char*  doh_method      = DEFAULT_DOH_METHOD;
+    const char*  local_suppress  = 0;
 
     sock_family     = AF_UNSPEC;
     server_port     = 0;
@@ -248,7 +272,7 @@ setup(int argc, char** argv)
     perf_opt_add('f', perf_opt_string, "family",
         "address family of DNS transport, inet or inet6", "any",
         &family);
-    perf_opt_add('M', perf_opt_string, "mode", "set transport mode: udp, tcp or dot", "udp", &_mode);
+    perf_opt_add('M', perf_opt_string, "mode", "set transport mode: udp, tcp, dot or doh", "udp", &_mode);
     perf_opt_add('s', perf_opt_string, "server_addr",
         "the server to query", DEFAULT_SERVER_NAME, &server_name);
     perf_opt_add('p', perf_opt_port, "port",
@@ -270,6 +294,8 @@ setup(int argc, char** argv)
         &bufsize);
     perf_opt_add('e', perf_opt_boolean, NULL,
         "enable EDNS 0", NULL, &edns);
+    perf_opt_add('E', perf_opt_string, "code:value",
+        "send EDNS option", NULL, &edns_option_str);
     perf_opt_add('D', perf_opt_boolean, NULL,
         "set the DNSSEC OK bit (implies EDNS)", NULL, &dnssec);
     perf_opt_add('y', perf_opt_string, "[alg:]name:secret",
@@ -305,6 +331,12 @@ setup(int argc, char** argv)
     perf_opt_add('R', perf_opt_boolean, NULL, "reopen datafile on end, allow for infinit use of it", NULL, &reopen_datafile);
     perf_opt_add('F', perf_opt_zpint, "fall_behind", "the maximum number of queries that is allowed to fall behind, zero to disable",
         stringify(DEFAULT_MAX_FALL_BEHIND, 0), &max_fall_behind);
+    perf_long_opt_add("doh-uri", perf_opt_string, "doh_uri",
+        "the URI to use for DNS-over-HTTPS", DEFAULT_DOH_URI, &doh_uri);
+    perf_long_opt_add("doh-method", perf_opt_string, "doh_method",
+        "the HTTP method to use for DNS-over-HTTPS: GET or POST", DEFAULT_DOH_METHOD, &doh_method);
+    perf_long_opt_add("suppress", perf_opt_string, "message[,message,...]",
+        "suppress messages/warnings, see dnsperf(1) man-page for list of message types", NULL, &local_suppress);
 
     perf_opt_parse(argc, argv);
 
@@ -312,13 +344,33 @@ setup(int argc, char** argv)
         perf_log_tostdout();
     }
 
+    suppress = perf_opt_parse_suppress(local_suppress);
+
     if (_mode != 0)
         mode = perf_net_parsemode(_mode);
 
     if (!server_port) {
-        server_port = mode == sock_dot ? DEFAULT_SERVER_DOT_PORT : DEFAULT_SERVER_PORT;
+        switch (mode) {
+        case sock_doh:
+            server_port = DEFAULT_SERVER_DOH_PORT;
+            break;
+        case sock_dot:
+            server_port = DEFAULT_SERVER_DOT_PORT;
+            break;
+        default:
+            server_port = DEFAULT_SERVER_PORT;
+            break;
+        }
     }
 
+    if (doh_uri) {
+        perf_net_doh_parse_uri(doh_uri);
+    }
+    if (doh_method) {
+        perf_net_doh_parse_method(doh_method);
+    }
+    perf_net_doh_set_max_concurrent_streams(max_outstanding);
+
     if (max_outstanding > nsocks * DEFAULT_MAX_OUTSTANDING)
         perf_log_fatal("number of outstanding packets (%u) must not "
                        "be more than 64K per client",
@@ -350,23 +402,32 @@ setup(int argc, char** argv)
         perf_datafile_setmaxruns(input, -1);
     }
 
-    if (dnssec)
+    if (dnssec || edns_option_str)
         edns = true;
 
     if (tsigkey_str != NULL)
         tsigkey = perf_tsig_parsekey(tsigkey_str);
 
+    if (edns_option_str != NULL)
+        edns_option = perf_edns_parseoption(edns_option_str);
+
+    traffic_time = ramp_time + sustain_time;
+    end_time     = traffic_time + wait_time;
+
+    n_buckets = (traffic_time + bucket_interval - 1) / bucket_interval;
+    buckets   = init_buckets(n_buckets);
+
+    time_now              = perf_get_time();
+    time_of_program_start = time_now;
+
     if (!(socks = calloc(nsocks, sizeof(*socks)))) {
         perf_log_fatal("out of memory");
     }
     for (i = 0; i < nsocks; i++) {
-        socks[i] = perf_net_opensocket(mode, &server_addr, &local_addr, i, bufsize);
+        socks[i] = perf_net_opensocket(mode, &server_addr, &local_addr, i, bufsize, (void*)(intptr_t)i, perf__net_sent, perf__net_event);
         if (!socks[i]) {
             perf_log_fatal("perf_net_opensocket(): no socket returned, out of memory?");
         }
-        socks[i]->data  = (void*)(intptr_t)i;
-        socks[i]->sent  = perf__net_sent;
-        socks[i]->event = perf__net_event;
     }
 }
 
@@ -376,10 +437,15 @@ cleanup(void)
     unsigned int i;
 
     perf_datafile_close(&input);
-    for (i = 0; i < nsocks; i++)
+    for (i = 0; i < nsocks; i++) {
+        perf_net_stats_compile(mode, socks[i]);
         (void)perf_net_close(socks[i]);
+    }
     close(dummypipe[0]);
     close(dummypipe[1]);
+
+    if (edns_option)
+        perf_edns_destroyoption(&edns_option);
 }
 
 /* Find the ramp_bucket for queries sent at time "when" */
@@ -405,11 +471,18 @@ static void perf__net_event(struct perf_net_socket* sock, perf_socket_event_t ev
     ramp_bucket* b = find_bucket(time_now);
 
     switch (event) {
-    case perf_socket_event_reconnect:
-        num_reconnections++;
-    case perf_socket_event_connect:
+    case perf_socket_event_reconnected:
+    case perf_socket_event_connected:
         b->connections++;
         b->conn_latency_sum += elapsed_time / (double)MILLION;
+        break;
+
+    case perf_socket_event_reconnecting:
+        num_reconnections++;
+        break;
+
+    default:
+        break;
     }
 }
 
@@ -481,23 +554,7 @@ print_statistics(void)
     }
     printf("  Maximum throughput:   %.6lf qps\n", max_throughput);
     printf("  Lost at that point:   %.2f%%\n", loss_at_max_throughput);
-}
-
-static ramp_bucket*
-init_buckets(int n)
-{
-    ramp_bucket* p;
-    int          i;
-
-    if (!(p = calloc(n, sizeof(*p)))) {
-        perf_log_fatal("out of memory");
-        return 0; // fix clang scan-build
-    }
-    for (i = 0; i < n; i++) {
-        p[i].queries = p[i].responses = p[i].failures = 0;
-        p[i].latency_sum                              = 0.0;
-    }
-    return p;
+    printf("\n");
 }
 
 /*
@@ -524,11 +581,11 @@ do_one_line(perf_buffer_t* lines, perf_buffer_t* msg)
     while (q->is_inprogress) {
         if (perf_net_sockready(socks[sock], dummypipe[0], TIMEOUT_CHECK_TIME) == -1) {
             if (errno == EINPROGRESS) {
-                if (verbose) {
+                if (verbose && !suppress.congestion) {
                     perf_log_warning("network congested, packet sending in progress");
                 }
             } else {
-                if (verbose) {
+                if (verbose && !suppress.sockready) {
                     char __s[256];
                     perf_log_warning("failed to check socket readiness: %s", perf_strerror_r(errno, __s, sizeof(__s)));
                 }
@@ -553,16 +610,16 @@ do_one_line(perf_buffer_t* lines, perf_buffer_t* msg)
 
     switch (perf_net_sockready(socks[sock], dummypipe[0], TIMEOUT_CHECK_TIME)) {
     case 0:
-        if (verbose) {
+        if (verbose && !suppress.sockready) {
             perf_log_warning("failed to send packet: socket %d not ready", sock);
         }
         return (PERF_R_FAILURE);
     case -1:
         if (errno == EINPROGRESS) {
-            if (verbose) {
+            if (verbose && !suppress.congestion) {
                 perf_log_warning("network congested, packet sending in progress");
             }
-        } else {
+        } else if (!suppress.sockready) {
             perf_log_warning("failed to send packet: socket %d not ready", sock);
         }
         return (PERF_R_FAILURE);
@@ -579,7 +636,7 @@ do_one_line(perf_buffer_t* lines, perf_buffer_t* msg)
     perf_buffer_clear(msg);
     result = perf_dns_buildrequest(&used, qid,
         edns, dnssec, false,
-        tsigkey, 0,
+        tsigkey, edns_option,
         msg);
     if (result != PERF_R_SUCCESS)
         return (result);
@@ -592,12 +649,12 @@ do_one_line(perf_buffer_t* lines, perf_buffer_t* msg)
             &server_addr.sa.sa, server_addr.length)
         < 1) {
         if (errno == EINPROGRESS) {
-            if (verbose) {
+            if (verbose && !suppress.congestion) {
                 perf_log_warning("network congested, packet sending in progress");
             }
             q->is_inprogress = true;
         } else {
-            if (verbose) {
+            if (verbose && !suppress.sendfailed) {
                 char __s[256];
                 perf_log_warning("failed to send packet: %s", perf_strerror_r(errno, __s, sizeof(__s)));
             }
@@ -652,7 +709,7 @@ try_process_response(unsigned int sockindex)
 
     if (perf_net_sockready(socks[sockindex], dummypipe[0], TIMEOUT_CHECK_TIME) == -1) {
         if (errno != EINPROGRESS) {
-            if (verbose) {
+            if (verbose && !suppress.sockready) {
                 char __s[256];
                 perf_log_warning("failed to check socket readiness: %s", perf_strerror_r(errno, __s, sizeof(__s)));
             }
@@ -758,6 +815,7 @@ int main(int argc, char** argv)
     switch (mode) {
     case sock_tcp:
     case sock_dot:
+    case sock_doh:
         // block SIGPIPE for TCP/DOT mode, if connection is closed it will generate a signal
         perf_os_blocksignal(SIGPIPE, true);
         break;
@@ -770,15 +828,6 @@ int main(int argc, char** argv)
     max_packet_size = edns ? MAX_EDNS_PACKET : MAX_UDP_PACKET;
     perf_buffer_init(&msg, outpacket_buffer, max_packet_size);
 
-    traffic_time = ramp_time + sustain_time;
-    end_time     = traffic_time + wait_time;
-
-    n_buckets = (traffic_time + bucket_interval - 1) / bucket_interval;
-    buckets   = init_buckets(n_buckets);
-
-    time_now              = perf_get_time();
-    time_of_program_start = time_now;
-
     printf("[Status] Command line: %s", progname);
     for (i = 1; i < argc; i++) {
         printf(" %s", argv[i]);
@@ -878,7 +927,9 @@ end_loop:
 
     fclose(plotf);
     print_statistics();
+    perf_net_stats_init(mode);
     cleanup();
+    perf_net_stats_print(mode);
 #if OPENSSL_VERSION_NUMBER < 0x10100000L
     ERR_free_strings();
 #endif
